import pathlib

import falcon
import falcon.asgi
from falcon.asgi.request import Request
from falcon.asgi.response import Response

from .chat import Chat
from .hub import Events
from .hub import Hub
from .ping import Pong

HERE = pathlib.Path(__file__).resolve().parent
STATIC = HERE.parent / 'static'


def create_app() -> falcon.asgi.App[Request, Response]:
    app = falcon.asgi.App()

    # NOTE(vytas): E2E tests run Uvicorn, and the latest versions support ASGI
    #   HTTP/WSspec ver 2.4, so buffering on our side should not be needed.
    app.ws_options.max_receive_queue = 0

    hub = Hub()
    app.add_route('/ping', Pong())
    app.add_route('/sse', Events(hub))
    app.add_route('/ws/{name}', Chat(hub))

    app.add_static_route('/static', STATIC)

    return app



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from uuid import UUID

import falcon
import falcon.asgi
import falcon.testing


@dataclass
class RichContext:
    userid: UUID | None = None
    role: str = 'anonymous'
    comment: str = 'no comment'


class FancyRequest(falcon.Request):
    context_type = RichContext


class FancyResponse(falcon.Response):
    context_type = RichContext

    # NOTE(vytas): the `type: ignore` exemption is currently required if it is
    #   desirable to actually check typing of context attributes. See also:
    #   https://falcon.readthedocs.io/en/latest/api/typing.html#known-limitations
    context: RichContext  # type: ignore[assignment]


class FancyAsyncRequest(falcon.asgi.Request):
    context_type = RichContext


class FancyAsyncResponse(falcon.asgi.Response):
    context_type = RichContext

    # NOTE(vytas): the `type: ignore` exemption is currently required if it is
    #   desirable to actually check typing of context attributes. See also:
    #   https://falcon.readthedocs.io/en/latest/api/typing.html#known-limitations
    context: RichContext  # type: ignore[assignment]


USERS = {
    'am9objoxMjM0': ('user', UUID('51e4b478-3825-4e46-9fd7-be7b61d616dc')),
    'dnl0YXM6MTIz': ('admin', UUID('5e50d2c4-1c52-42c7-b4c5-879d9bd390ee')),
}


def fancy_error_serializer(
    req: FancyRequest, resp: falcon.Response, ex: falcon.HTTPError
) -> None:
    resp.content_type = falcon.MEDIA_JSON
    resp.media = ex.to_dict()
    resp.media.update(fancy=True, asgi=False)


def fancy_asgi_error_serializer(
    req: FancyAsyncRequest, resp: falcon.asgi.Response, ex: falcon.HTTPError
) -> None:
    resp.content_type = falcon.MEDIA_JSON
    resp.media = ex.to_dict()
    resp.media.update(fancy=True, asgi=True)


def _process_auth(req: falcon.Request, resp: falcon.Response) -> None:
    if req.method == 'OPTIONS':
        return

    if req.auth:
        for key, user_role in USERS.items():
            if req.auth == f'Basic {key}':
                req.context.role, req.context.userid = user_role
                break
        else:
            raise falcon.HTTPUnauthorized()


class AuthMiddlewareFancyRequest:
    def process_request(self, req: FancyRequest, resp: falcon.Response) -> None:
        _process_auth(req, resp)

    async def process_request_async(
        self, req: FancyAsyncRequest, resp: falcon.asgi.Response
    ) -> None:
        _process_auth(req, resp)


class AuthMiddlewareFancyBoth:
    def process_request(self, req: FancyRequest, resp: FancyResponse) -> None:
        _process_auth(req, resp)

        # NOTE(vytas): Unlike req.context, resp.context.comment is type checked,
        #   try misspelling it or using with an incompatible type.
        resp.context.comment = 'fancy req/resp'

    async def process_request_async(
        self, req: FancyAsyncRequest, resp: FancyAsyncResponse
    ) -> None:
        _process_auth(req, resp)
        resp.context.comment = 'fancy req/resp'


def _sink_impl(req: falcon.Request, resp: falcon.Response) -> None:
    userid: str | None = str(req.context.userid) if req.context.userid else None
    resp.media = {'role': req.context.role, 'userid': userid}
    if req.path == '/not-found':
        raise falcon.HTTPNotFound()


def sink_fancy_req(req: FancyRequest, resp: falcon.Response, **kwargs: Any) -> None:
    _sink_impl(req, resp)


def sink_fancy_both(req: FancyRequest, resp: FancyResponse, **kwargs: Any) -> None:
    _sink_impl(req, resp)
    resp.context.comment += ' (sink)'
    resp.media.update(comment=resp.context.comment)


async def sink_fancy_async_req(
    req: FancyAsyncRequest, resp: falcon.asgi.Response | None, **kwargs: Any
) -> None:
    if resp is not None:
        _sink_impl(req, resp)


async def sink_fancy_async_both(
    req: FancyAsyncRequest, resp: FancyAsyncResponse | None, **kwargs: Any
) -> None:
    if resp is not None:
        _sink_impl(req, resp)
        resp.context.comment += ' (sink)'
        resp.media.update(comment=resp.context.comment)


# NOTE(vytas): We don't use fixtures here because that is hard to marry to strict
#   type checking, Mypy complains that "Untyped decorator makes function untyped".
def test_app_fancy_req() -> None:
    app = falcon.App(request_type=FancyRequest)

    app.add_middleware(AuthMiddlewareFancyRequest())
    app.add_sink(sink_fancy_req)
    app.set_error_serializer(fancy_error_serializer)

    _exercise_app(app)


def test_app_fancy_both() -> None:
    app = falcon.App(request_type=FancyRequest, response_type=FancyResponse)

    app.add_middleware(AuthMiddlewareFancyBoth())
    app.add_sink(sink_fancy_both)
    app.set_error_serializer(fancy_error_serializer)

    _exercise_app(app)


def test_app_fancy_async_req() -> None:
    app = falcon.asgi.App(request_type=FancyAsyncRequest)

    app.add_middleware(AuthMiddlewareFancyRequest())
    app.add_sink(sink_fancy_async_req)
    app.set_error_serializer(fancy_asgi_error_serializer)

    _exercise_app(app)


def test_app_fancy_async_both() -> None:
    app = falcon.asgi.App(
        request_type=FancyAsyncRequest, response_type=FancyAsyncResponse
    )

    app.add_middleware(AuthMiddlewareFancyBoth())
    app.add_sink(sink_fancy_async_both)
    app.set_error_serializer(fancy_asgi_error_serializer)

    _exercise_app(app)


def _exercise_app(app: falcon.App[Any, Any]) -> None:
    client = falcon.testing.TestClient(app)

    result1 = client.get()
    assert result1.status_code == 200
    assert result1.json in (
        {
            'role': 'anonymous',
            'userid': None,
        },
        {
            'comment': 'fancy req/resp (sink)',
            'role': 'anonymous',
            'userid': None,
        },
    )

    result2 = client.get(headers={'Authorization': 'ApiKey 123'})
    assert result2.status_code == 401

    result3 = client.get(headers={'Authorization': 'Basic am9objoxMjM0'})
    assert result3.status_code == 200
    assert result3.json in (
        {
            'role': 'user',
            'userid': '51e4b478-3825-4e46-9fd7-be7b61d616dc',
        },
        {
            'comment': 'fancy req/resp (sink)',
            'role': 'user',
            'userid': '51e4b478-3825-4e46-9fd7-be7b61d616dc',
        },
    )

    result4 = client.get('/not-found')
    assert result4.status_code == 404
    assert result4.json == {
        'asgi': isinstance(client.app, falcon.asgi.App),
        'fancy': True,
        'title': '404 Not Found',
    }



import asyncio
from collections import deque
from contextlib import nullcontext
import functools
import os

import pytest

import falcon
from falcon import media
from falcon import testing
from falcon.asgi import App
from falcon.asgi.ws import _WebSocketState as ServerWebSocketState
from falcon.asgi.ws import WebSocket
from falcon.asgi.ws import WebSocketOptions
from falcon.testing.helpers import _WebSocketState as ClientWebSocketState
from falcon.util.deprecation import DeprecatedWarning

try:
    import cbor2
except ImportError:
    cbor2 = None  # type: ignore[assignment]


try:
    import msgpack
except ImportError:
    msgpack = None


try:
    import rapidjson
except ImportError:
    rapidjson = None  # type: ignore[assignment]


# NOTE(kgriffs): We do not use codes defined in the framework because we
#   want to verify that the correct value is being used.
class CloseCode:
    NORMAL = 1000
    SERVER_ERROR = 1011
    FORBIDDEN = 3403
    NOT_FOUND = 3404


class EventType:
    WS_CONNECT = 'websocket.connect'
    WS_ACCEPT = 'websocket.accept'
    WS_RECEIVE = 'websocket.receive'
    WS_SEND = 'websocket.send'
    WS_DISCONNECT = 'websocket.disconnect'
    WS_CLOSE = 'websocket.close'


@pytest.fixture
def conductor():
    app = falcon.asgi.App()
    return testing.ASGIConductor(app)


@pytest.mark.parametrize('path', ['/ws/yes', '/ws/no'])
async def test_ws_not_accepted(path, conductor):
    class SomeResource:
        def __init__(self):
            self.called = False
            self.caught_unsupported_error = False
            self.caught_receive_not_allowed = False
            self.caught_operation_not_allowed = False

        async def on_websocket(self, req, ws, explicit):
            try:
                req.stream
            except falcon.UnsupportedError:
                self.caught_unsupported_error = True

            self.called = True

            try:
                await ws.receive_text()
            except falcon.OperationNotAllowed:
                self.caught_receive_not_allowed = True

            if explicit == 'yes':
                await ws.close()

                # NOTE(kgriffs): This verifies that closing
                #   an already-closed WebSocket is a no-op.
                await ws.close()

                try:
                    await ws.accept()
                except falcon.OperationNotAllowed:
                    self.caught_operation_not_allowed = True

    resource = SomeResource()
    conductor.app.add_route('/ws/{explicit}', resource)

    # NOTE(kgriffs): declaring the 'c' variable is not strictly necessary
    #   because it is just 'conductor', but it probably wastes fewer brain
    #   cycles to just use the same pattern whether or not the conductor
    #   object is pre-instantiated.
    async with conductor as c:
        assert c is conductor

        with pytest.raises(falcon.WebSocketDisconnected) as exc_info:
            async with c.simulate_ws(path):
                pass

        assert resource.called
        assert resource.caught_receive_not_allowed
        assert resource.caught_unsupported_error
        assert exc_info.value.code == CloseCode.FORBIDDEN

        if 'yes' in path:
            assert resource.caught_operation_not_allowed


@pytest.mark.slow
async def test_echo():  # noqa: C901
    consumer_sleep = 0.01
    producer_loop = 10
    producer_sleep_factor = consumer_sleep / (producer_loop / 2)

    class Resource:
        def __init__(self):
            self.caught_operation_not_allowed = False

        async def on_websocket(self, req, ws, p1, p2, injected):
            # NOTE(kgriffs): Normally the receiver task is not started
            #   until the websocket is started. But here we start it
            #   first in order to simulate a potential race condition
            #   that ASGIWebSocketSimulator._emit() guards against,
            #   in case it ever arises due to the way the target ASGI
            #   app may be implemented.
            ws._buffered_receiver.start()

            await asyncio.sleep(0)

            if ws.unaccepted:
                await ws.accept()

            try:
                await ws.accept()
            except falcon.OperationNotAllowed:
                self.caught_operation_not_allowed = True

            if ws.ready:
                await ws.send_text(f'{p1}:{p2}:{req.context.message}:{injected}')

            messages = deque()
            sink_task = asyncio.create_task(self._sink(ws, messages))

            while not sink_task.done():
                if not messages:
                    await asyncio.sleep(0)
                    continue

                try:
                    await ws.send_text(messages.popleft())
                except falcon.WebSocketDisconnected:
                    break

            sink_task.cancel()
            try:
                await sink_task
            except asyncio.CancelledError:
                pass

        async def _sink(self, ws, messages):
            while True:
                # NOTE(kgriffs): Throttle slightly to allow messages to
                #   fill up the buffer.
                await asyncio.sleep(consumer_sleep)

                try:
                    message = await ws.receive_text()
                except falcon.WebSocketDisconnected:
                    break

                if message != 'ignore':
                    messages.append(message)

    class MiddlewareA:
        async def process_resource_ws(self, req, ws, resource, params):
            assert isinstance(resource, Resource)
            assert isinstance(ws, falcon.asgi.WebSocket)
            params['injected'] = '42'

    class MiddlewareB:
        async def process_request_ws(self, req, ws):
            assert isinstance(ws, falcon.asgi.WebSocket)
            req.context.message = 'hello'

    resource = Resource()

    # NOTE(kgriffs): The two methods are split across different middleware
    #   classes so that we can test code paths that check for the existence
    #   of one WebSocket middleware method vs the other, and also so that
    #   we can make sure both the kwarg and the add_middleware() paths
    #   succeed.
    app = App(middleware=MiddlewareA())
    app.add_middleware(MiddlewareB())

    app.add_route('/{p1}/{p2}', resource)

    async with testing.ASGIConductor(app) as c:
        # NOTE(vytas): Using the websocket() alias.
        async with c.websocket('/v1/v2', headers={}) as ws:
            assert (await ws.receive_text()) == 'v1:v2:hello:42'

            for i in range(producer_loop):
                message = (
                    str(i) if i else ''
                )  # Test round-tripping the empty string as well

                for i in range(100):
                    await ws.send_text('ignore')

                # NOTE(kgriffs): For part of the time, cause the buffer on the
                #   server side to fill up, and for the remainder of the time
                #   for the buffer to be empty and wait on the client for
                #   additional messages.
                await asyncio.sleep(i * producer_sleep_factor)

                await ws.send_text(message)
                assert (await ws.receive_text()) == message

            await ws.close()

            assert ws.closed
            assert ws.close_code == CloseCode.NORMAL

    assert resource.caught_operation_not_allowed


async def test_path_not_found(conductor):
    async with conductor as c:
        with pytest.raises(falcon.WebSocketDisconnected) as exc_info:
            async with c.simulate_ws():
                pass

        assert exc_info.value.code == CloseCode.NOT_FOUND


@pytest.mark.parametrize('clear_error_handlers', [True, False])
async def test_responder_raises_unhandled_error(clear_error_handlers, conductor):
    class SomeResource:
        def __init__(self):
            self.called = False

        async def on_websocket(self, req, ws, generic_error):
            self.called = True

            if generic_error == 'true':
                raise Exception('Oh snap!')

            # NOTE(kgriffs): This may not make a lot of sense to do, but the
            #   framework will try to do something reasonable with it by
            #   rejecting the WebSocket handshake and setting the close code
            #   to 3000 + falcon.util.http_status_to_code(error.status)
            raise falcon.HTTPUnprocessableEntity()

    resource = SomeResource()
    conductor.app.add_route('/{generic_error}', resource)

    if clear_error_handlers:
        # NOTE(kgriffs): This is required in order to cover a certain branch condition.
        conductor.app._error_handlers.clear()

        async with conductor as c:
            with pytest.raises(Exception) as exc_info:
                async with c.simulate_ws('/true'):
                    pass

            assert resource.called
            assert str(exc_info.value) == 'Oh snap!'

            resource.called = False

            with pytest.raises(falcon.HTTPUnprocessableEntity):
                async with c.simulate_ws('/false'):
                    pass

            assert resource.called
    else:
        async with conductor as c:
            with pytest.raises(falcon.WebSocketServerError) as exc_info:
                async with c.simulate_ws('/true'):
                    pass

            assert resource.called
            assert exc_info.value.code == 1011

            resource.called = False

            with pytest.raises(falcon.WebSocketDisconnected) as exc_info:
                async with c.simulate_ws('/false'):
                    pass

            assert resource.called
            assert exc_info.value.code == 3422


@pytest.mark.parametrize('direction', ['send', 'receive'])
@pytest.mark.parametrize('explicit_close_client', [True, False])
@pytest.mark.parametrize('explicit_close_server', [True, False])
async def test_client_disconnect_early(  # noqa: C901
    direction,
    explicit_close_client,
    explicit_close_server,
    conductor,
):
    sample_data = os.urandom(64 * 1024)

    class Resource:
        def __init__(self):
            self.data_received = asyncio.Event()
            self.times_called = 0
            self.data = None
            self.ws_ready = None
            self.ws_state = None
            self.ws_closed = None
            self.ws_client_close_code = None

        async def on_websocket(self, req, ws):
            self.times_called += 1

            await ws.accept()

            while True:
                try:
                    if direction == 'send':
                        await ws.send_data(sample_data)

                        if explicit_close_server:
                            await ws.close(4099)
                    else:
                        self.data = await ws.receive_data()

                        if explicit_close_server:
                            # NOTE(kgriffs): We call ws.receive_data() again here in
                            #   order to test coverage of the logic that handles
                            #   the case of a closed connection while waiting on
                            #   more data.
                            recv_task = asyncio.create_task(ws.receive_data())
                            # Ensure recv_task() has a chance to get ahead
                            await asyncio.sleep(0)
                            ws_close = ws.close(4099)
                            ws_close = asyncio.create_task(ws_close)
                            await asyncio.wait([recv_task, ws_close])

                        self.data_received.set()

                except falcon.WebSocketDisconnected as ex:
                    self.ws_ready = ws.ready
                    self.ws_client_close_code = ex.code
                    self.ws_state = ws._state
                    self.ws_closed = ws.closed
                    break

    resource = Resource()
    conductor.app.add_route('/', resource)

    disconnect_raised = False

    async with conductor as c:
        try:
            async with c.simulate_ws('/') as ws:
                if direction == 'send':
                    received_data = await ws.receive_data()
                    assert received_data == sample_data
                else:
                    await ws.send_data(sample_data)
                    await resource.data_received.wait()
                    assert resource.data == sample_data

                # NOTE(vytas): When testing the case where the server
                #   explicitly closes the connection, try to receive some data
                #   before closing from the client side (and potentially
                #   winning the async race of which side closes first).
                if explicit_close_server:
                    await ws.receive_data()

                if explicit_close_client:
                    await ws.close(4042)

        except falcon.WebSocketDisconnected as ex:
            disconnect_raised = True
            assert explicit_close_server
            assert ex.code == 4099

        if direction == 'send' and explicit_close_server:
            assert disconnect_raised

        code_expected = CloseCode.NORMAL

        if explicit_close_server:
            code_expected = 4099
        elif explicit_close_client:
            code_expected = 4042

        assert resource.ws_client_close_code == code_expected
        assert ws.close_code == code_expected

    assert resource.times_called == 1
    assert resource.ws_state == ServerWebSocketState.CLOSED
    assert resource.ws_closed
    assert resource.ws_ready is False


@pytest.mark.parametrize('custom_text', [True, False])
@pytest.mark.parametrize('custom_data', [True, False])
@pytest.mark.skipif(msgpack is None, reason='msgpack is required for this test')
async def test_media(custom_text, custom_data, conductor):  # NOQA: C901
    # TODO(kgriffs): Refactor to reduce McCabe score

    sample_doc = {
        'answer': 42,
        'runes': b'\xe1\x9a\xa0\xe1\x9b\x87\xe1\x9a\xbb'.decode(),
        'ascii': 'hello world',
    }

    sample_doc_bin = sample_doc.copy()
    sample_doc_bin['bits'] = os.urandom(32)
    sample_doc_bin['array'] = [0, 1, 2]

    class Resource:
        def __init__(self):
            self.finished = asyncio.Event()
            self.docs_received = []
            self.deserialize_error_count = 0

        async def on_websocket(self, req, ws):
            try:
                await ws.accept()

                await ws.send_media(sample_doc)
                self.docs_received.append(await ws.receive_media())
                await ws.send_media(sample_doc, falcon.WebSocketPayloadType.TEXT)
                self.docs_received.append(await ws.receive_media())

                await ws.send_media(sample_doc_bin, falcon.WebSocketPayloadType.BINARY)
                self.docs_received.append(await ws.receive_media())

                for __ in range(3):
                    try:
                        await ws.receive_media()
                    except ValueError:
                        self.deserialize_error_count += 1
            finally:
                self.finished.set()

    app = conductor.app

    resource = Resource()
    app.add_route('/', resource)

    if custom_text:
        if rapidjson is None:
            pytest.skip('rapidjson is required for this test')

        # Let's say we want to use a faster JSON library. You could also use this
        #   pattern to add serialization support for custom types that aren't
        #   normally JSON-serializable out of the box.
        class RapidJSONHandler(media.TextBaseHandlerWS):
            def serialize(self, media: object) -> str:
                return rapidjson.dumps(media, ensure_ascii=False)

            # The raw TEXT payload will be passed as a Unicode string
            def deserialize(self, payload: str) -> object:
                return rapidjson.loads(payload)

        app.ws_options.media_handlers[falcon.WebSocketPayloadType.TEXT] = (
            RapidJSONHandler()
        )

    if custom_data:
        if cbor2 is None:
            pytest.skip('cbor2 is required for this test')

        class CBORHandler(media.BinaryBaseHandlerWS):
            def serialize(self, media: object) -> bytes:
                return cbor2.dumps(media)

            # The raw BINARY payload will be passed as a byte string
            def deserialize(self, payload: bytes) -> object:
                try:
                    return cbor2.loads(payload)
                except cbor2.CBORDecodeError as ex:
                    # NOTE(vytas): As of cbor2 6.0, CBORDecodeError is no
                    #   longer a subclass of ValueError, so we reraise to
                    #   signal an invalid payload (mirroring the pattern
                    #   documented in our WebSocket guide).
                    raise ValueError('error decoding CBOR payload') from ex

        app.ws_options.media_handlers[falcon.WebSocketPayloadType.BINARY] = (
            CBORHandler()
        )

    async with conductor as c:
        async with c.simulate_ws() as ws:
            for __ in range(2):
                doc = await ws.receive_json()
                await ws.send_json(doc)

            if custom_data:
                data = await ws.receive_data()
                cbor2.loads(data)  # NOTE(kgriffs): Validate serialization format
                await ws.send_data(data)
            else:
                doc = await ws.receive_msgpack()
                await ws.send_msgpack(doc)

            # NOTE(kgriffs): The first one will work, but we include it to
            #    ensure we aren't getting any false-positives.
            await ws.send_text('"DEADBEEF"')
            await ws.send_text('DEADBEEF')
            await ws.send_data(b'\xde\xad\xbe\xef')

            await resource.finished.wait()

    assert resource.deserialize_error_count == 2

    assert len(resource.docs_received) == 3
    assert resource.docs_received[0] == sample_doc

    expected = [sample_doc, sample_doc, sample_doc_bin]
    for doc, doc_expected in zip(resource.docs_received, expected):
        assert doc == doc_expected


@pytest.mark.parametrize('sample_data', [b'123', b'', b'\xe1\x9a\xa0\xe1', b'\0'])
async def test_send_receive_data(sample_data, conductor):
    class Resource:
        def __init__(self):
            self.error_count = 0

        async def on_websocket(self, req, ws):
            await ws.accept()

            await ws.send_data(await ws.receive_data())
            await ws.send_data(bytearray(await ws.receive_data()))
            await ws.send_data(memoryview(await ws.receive_data()))

            for c in (ws.receive_data, ws.receive_media):
                for __ in range(2):
                    try:
                        await c()
                    except falcon.PayloadTypeError:
                        self.error_count += 1

    resource = Resource()
    conductor.app.add_route('/', resource)

    async with conductor as c:
        async with c.simulate_ws() as ws:
            await ws.send_data(sample_data)
            await ws.send_data(bytearray(sample_data))
            await ws.send_data(memoryview(sample_data))

            for __ in range(2):
                ws._collected_client_events.append({'type': EventType.WS_RECEIVE})
                ws._collected_client_events.append(
                    {'type': EventType.WS_RECEIVE, 'bytes': None}
                )

            for __ in range(3):
                assert (await ws.receive_data()) == sample_data

    assert resource.error_count == 4


@pytest.mark.parametrize(
    'subprotocols',
    [
        ['SIS508'],
        ('DEADBEEF', 'SIS508'),
        [],
        None,
    ],
)
async def test_subprotocol(subprotocols, conductor):
    class Resource:
        def __init__(self):
            self.test_complete = asyncio.Event()

        async def on_websocket(self, req, ws):
            subs = ws.subprotocols
            assert isinstance(subs, tuple)

            selected = subs[0] if subs else None
            await ws.accept(subprotocol=selected)

            await self.test_complete.wait()

    resource = Resource()
    conductor.app.add_route('/', resource)

    async with conductor as c:
        async with c.simulate_ws(subprotocols=subprotocols) as ws:
            try:
                if not subprotocols:
                    assert ws.subprotocol is None
                else:
                    assert ws.subprotocol == subprotocols[0]
            finally:
                resource.test_complete.set()


@pytest.mark.parametrize(
    'headers',
    [
        None,
        [],
        iter([]),
        iter([iter(['X-Name', 'Value']), ('NAME', 'Value'), ('NAME', 'Value')]),
        (['X-Name', 'Value'], ('NAME', 'Value'), ('NAME', 'Value')),
        {'NAME': 'value'},
        {'NAME': 'value', 'NameToo': 'ValueToo'},
    ],
)
async def test_accept_with_headers(headers, conductor):
    class Resource:
        def __init__(self):
            self.test_complete = asyncio.Event()

        async def on_websocket(self, req, ws):
            await ws.accept(headers=headers)
            await self.test_complete.wait()

    resource = Resource()
    conductor.app.add_route('/', resource)

    async with conductor as c:
        async with c.simulate_ws() as ws:
            try:
                if not headers:
                    assert ws.headers is None
                else:
                    headers_in = headers
                    if isinstance(headers, dict):
                        headers_in = headers.items()

                    for (name_in, value_in), (name_out, value_out) in zip(
                        headers_in, ws.headers
                    ):
                        name_out = name_out.decode()
                        value_out = value_out.decode()

                        assert name_out.islower()
                        assert name_in.lower() == name_out
                        assert value_in == value_out

            finally:
                resource.test_complete.set()


@pytest.mark.parametrize(
    'headers',
    [
        [['sec-websocket-protocol', 'SIS508']],
        [['non-ascii', b'\xe1\x9a\xa0\xe1\x9b\x87\xe1\x9a\xbb'.decode()]],
        {b'\xe1\x9a\xa0\xe1\x9b\x87\xe1\x9a\xbb'.decode(): 'non-ascii'},
    ],
)
async def test_accept_with_bad_headers(headers, conductor):
    class Resource:
        def __init__(self):
            self.raised_error = None

        async def on_websocket(self, req, ws):
            try:
                assert ws.supports_accept_headers
                await ws.accept(headers=headers)
            except Exception as ex:
                self.raised_error = ex

    resource = Resource()
    conductor.app.add_route('/', resource)

    async with conductor as c:
        try:
            async with c.simulate_ws():
                pass
        except falcon.WebSocketDisconnected:
            pass

    assert isinstance(resource.raised_error, ValueError)


async def test_accept_with_headers_not_supported(conductor):
    class Resource:
        def __init__(self):
            self.raised_error = None

        async def on_websocket(self, req, ws):
            try:
                assert not ws.supports_accept_headers
                await ws.accept(headers={'name': 'value'})
            except Exception as ex:
                self.raised_error = ex

    resource = Resource()
    conductor.app.add_route('/', resource)

    async with conductor as c:
        try:
            async with c.simulate_ws(spec_version='2.0'):
                pass
        except falcon.WebSocketDisconnected:
            pass

    assert isinstance(resource.raised_error, falcon.OperationNotAllowed)


async def test_missing_ws_handler(conductor):
    class Resource:
        async def on_get(self, req, resp):
            pass

    conductor.app.add_route('/', Resource())

    async with conductor as c:
        with pytest.raises(falcon.WebSocketHandlerNotFound):
            async with c.simulate_ws():
                pass


async def test_unexpected_param(conductor):
    class Resource:
        async def on_websocket(self, req, ws):
            pass

    conductor.app.add_route('/{id}', Resource())

    async with conductor as c:
        with pytest.raises(falcon.WebSocketServerError):
            async with c.simulate_ws('/DEADBEEF'):
                pass


@pytest.mark.parametrize(
    'subprotocol',
    [
        b'DEADBEEF',
        ['SIS508'],
        [],
        tuple(),
        {},
        'OK',  # control
    ],
)
async def test_subprotocol_bad_type(subprotocol, conductor):
    class Resource:
        def __init__(self):
            self.test_complete = asyncio.Event()

        async def on_websocket(self, req, ws):
            await ws.accept(subprotocol=subprotocol)

    resource = Resource()
    conductor.app.add_route('/', resource)

    async with conductor as c:
        if isinstance(subprotocol, str):
            async with c.simulate_ws():
                pass
        else:
            with pytest.raises(falcon.WebSocketServerError):
                async with c.simulate_ws():
                    pass


async def test_send_receive_wrong_type(conductor):
    class Resource:
        def __init__(self):
            self.error_count = 0

        async def on_websocket(self, req, ws):
            await ws.accept()

            try:
                await ws.receive_data()
            except falcon.PayloadTypeError:
                self.error_count += 1

            try:
                await ws.receive_text()
            except falcon.PayloadTypeError:
                self.error_count += 1

            try:
                await ws.send_text(b'')
            except TypeError:
                self.error_count += 1

            try:
                await ws.send_data('')
            except TypeError:
                self.error_count += 1

            await ws.send_text('text')
            await ws.send_data(b'data')

    resource = Resource()
    conductor.app.add_route('/', resource)

    async with conductor as c:
        async with c.simulate_ws() as ws:
            with pytest.raises(TypeError):
                await ws.send_text(b'')

            with pytest.raises(TypeError):
                await ws.send_data('')

            await ws.send_text('text')
            await ws.send_data(b'data')

            with pytest.raises(falcon.PayloadTypeError):
                await ws.receive_data()

            with pytest.raises(falcon.PayloadTypeError):
                await ws.receive_text()

    assert resource.error_count == 4


@pytest.mark.parametrize(
    'options_code',
    [999, 100, 0, -1, 1004, 1005, 1006, 1015, 1016, 1017, 1050, 1099, 'NaN'],
)
async def test_client_disconnect_uncaught_error(options_code, conductor):
    class Resource:
        def __init__(self):
            self.disconnected = False

        async def on_websocket(self, req, ws):
            await ws.accept()
            while True:
                try:
                    await ws.receive_text()
                except falcon.WebSocketDisconnected:
                    self.disconnected = True
                    raise

    resource = Resource()
    conductor.app.add_route('/', resource)
    conductor.app.ws_options.error_close_code = options_code

    async with conductor as c:
        try:
            async with c.simulate_ws('/') as ws:
                await ws.send_text('hello')
        except ValueError:
            assert options_code == 'NaN'

    assert resource.disconnected
    assert ws.close_code == CloseCode.NORMAL


def test_mw_methods_must_be_coroutines():
    class MiddlewareA:
        def process_resource_ws(self, req, ws, resource, params):
            pass

    class MiddlewareB:
        def process_request_ws(self, req, ws):
            pass

    app = App()

    for mw in [MiddlewareA(), MiddlewareB()]:
        with pytest.raises(falcon.CompatibilityError):
            app.add_middleware(mw)

        with pytest.raises(falcon.CompatibilityError):
            App(middleware=mw)


@pytest.mark.parametrize('version', ['1.9', '20.5', '3.0', '3.1'])
async def test_bad_spec_version(version, conductor):
    async with conductor as c:
        with pytest.raises(falcon.UnsupportedScopeError):
            async with c.simulate_ws('/', spec_version=version):
                pass


@pytest.mark.parametrize('version', ['1.0', '1'])
async def test_bad_http_version(version, conductor):
    async with conductor as c:
        with pytest.raises(falcon.UnsupportedError):
            async with c.simulate_ws('/', http_version=version):
                pass


@pytest.mark.parametrize('version', ['2.1', '2.3', '2.10.3'])
async def test_bad_first_event(version):
    app = App()

    scope = testing.create_scope_ws(spec_version=version)

    ws = testing.ASGIWebSocketSimulator()
    wrapped_emit = ws._emit

    async def _emit():
        if ws._state == ClientWebSocketState.CONNECT:
            ws._state = ClientWebSocketState.HANDSHAKE
            return {'type': EventType.WS_SEND}

        return await wrapped_emit(ws)

    ws._emit = _emit

    # NOTE(kgriffs): If there is a bad first event, the framework should
    #   just bail out early and close the request...
    await app(scope, ws._emit, ws._collect)

    assert ws.closed
    assert ws.close_code == CloseCode.SERVER_ERROR
    if version != '2.1':
        assert ws.close_reason == 'Internal Server Error'
    else:
        assert ws.close_reason == ''


async def test_missing_http_version():
    app = App()

    scope = testing.create_scope_ws()
    del scope['http_version']

    ws = testing.ASGIWebSocketSimulator()

    # NOTE(kgriffs): As long as this does not raise, we know the http
    #   version defaulted OK.
    await app(scope, ws._emit, ws._collect)


async def test_missing_spec_version():
    app = App()

    scope = testing.create_scope_ws()
    del scope['asgi']['spec_version']

    ws = testing.ASGIWebSocketSimulator()

    # NOTE(kgriffs): As long as this does not raise, we know the spec
    #   version defaulted OK.
    await app(scope, ws._emit, ws._collect)


async def test_translate_webserver_error(conductor):
    class Resource:
        def __init__(self):
            self.error_count = 0
            self.test_complete = asyncio.Event()

        async def on_websocket(self, req, ws):
            await ws.accept()

            async def raise_disconnect(event):
                raise Exception('Disconnected with code = 1000 (OK)')

            async def raise_io_error(event, cause=''):
                class ConnectionClosed(RuntimeError):
                    pass

                class ClientDisconnected(OSError):
                    pass

                if cause:
                    raise ClientDisconnected() from ConnectionClosed(cause)
                raise ClientDisconnected()

            async def raise_protocol_mismatch(event):
                raise Exception('protocol accepted must be from the list')

            async def raise_other(event):
                raise RuntimeError()

            # TODO(vytas): It would be nice to somehow refactor this test not
            #   to operate on the private members of WebSocket.
            #   But, OTOH, it is quite useful as it is.
            _asgi_send = ws._asgi_send
            _original_state = ws._state

            ws._asgi_send = raise_other
            ws._state = _original_state
            try:
                await ws.send_data(b'123')
            except Exception:
                self.error_count += 1

            ws._asgi_send = raise_protocol_mismatch
            ws._state = _original_state
            try:
                await ws.send_data(b'123')
            except ValueError:
                self.error_count += 1

            ws._asgi_send = raise_disconnect
            ws._state = _original_state
            try:
                await ws.send_data(b'123')
            except falcon.WebSocketDisconnected:
                self.error_count += 1

            ws._asgi_send = raise_io_error
            ws._state = _original_state
            try:
                await ws.send_data(b'123')
            except falcon.WebSocketDisconnected:
                self.error_count += 1

            ws._asgi_send = functools.partial(raise_io_error, cause='bork3d pipe')
            ws._state = _original_state
            try:
                await ws.send_data(b'123')
            except falcon.WebSocketDisconnected:
                self.error_count += 1

            ws._asgi_send = functools.partial(raise_io_error, cause='received 1001')
            ws._state = _original_state
            try:
                await ws.send_data(b'123')
            except falcon.WebSocketDisconnected:
                self.error_count += 1
            assert ws._close_code == 1001

            ws._asgi_send = _asgi_send
            ws._state = _original_state

            self.test_complete.set()

    resource = Resource()
    conductor.app.add_route('/', resource)

    async with conductor as c:
        async with c.simulate_ws():
            await resource.test_complete.wait()

    assert resource.error_count == 6


def test_ws_base_not_implemented():
    th = media.TextBaseHandlerWS()

    with pytest.raises(NotImplementedError):
        th.serialize({})

    with pytest.raises(NotImplementedError):
        th.deserialize('')

    bh = media.BinaryBaseHandlerWS()

    with pytest.raises(NotImplementedError):
        bh.serialize({})

    with pytest.raises(NotImplementedError):
        bh.deserialize(b'')


@pytest.mark.slow
async def test_ws_context_timeout(conductor):
    class Resource:
        async def on_websocket(self, req, ws):
            await asyncio.sleep(5.1)  # Anything longer than 5 is sufficient

    conductor.app.add_route('/', Resource())

    async with conductor as c:
        with pytest.raises(asyncio.TimeoutError):
            async with c.simulate_ws():
                pass


async def test_ws_simulator_client_require_accepted(conductor):
    class Resource:
        async def on_websocket(self, req, ws):
            await asyncio.sleep(5.1)  # Anything longer than 5 is sufficient

    conductor.app.add_route('/', Resource())

    async with conductor as c:
        context = c.simulate_ws()
        ws = context._ws

        assert not ws.ready

        m = 'not yet been accepted'
        with pytest.raises(falcon.OperationNotAllowed, match=m):
            await ws.receive_text()


async def test_ws_simulator_collect_edge_cases(conductor):
    class Resource:
        pass

    conductor.app.add_route('/', Resource())

    async with conductor as c:
        context = c.simulate_ws(spec_version='2.3')
        ws = context._ws

        m = 'must receive the first websocket.connect'
        with pytest.raises(falcon.OperationNotAllowed, match=m):
            await ws._collect({'type': 'test'})

        event = await ws._emit()
        assert event['type'] == EventType.WS_CONNECT

        m = 'before sending any other event types'
        with pytest.raises(falcon.OperationNotAllowed, match=m):
            await ws._collect({'type': EventType.WS_SEND})

        await ws.close()

        # NOTE(vytas): The collector should start raising an instance of
        #   OSError from now on per the ASGI WS spec ver 2.4+.
        for __ in range(10):
            with pytest.raises(OSError):
                await ws._collect({'type': EventType.WS_SEND})

        assert not ws._collected_server_events

        event = await ws._emit()
        assert event['type'] == EventType.WS_DISCONNECT

        m = 'websocket.disconnect event has already been emitted'
        with pytest.raises(falcon.OperationNotAllowed, match=m):
            event = await ws._emit()


@pytest.mark.slow
async def test_ws_responder_never_ready(conductor, monkeypatch):
    async def noop_close(obj, code=None):
        pass

    class SleepyResource:
        async def on_websocket(self, req, ws):
            for i in range(10):
                await asyncio.sleep(0.001)

    conductor.app.add_route('/', SleepyResource())

    # NOTE(vytas): It seems that it is hard to impossible to hit the second
    #   `await ready_waiter` of the _WSContextManager on CPython 3.12 due to
    #   different async code optimizations, so we mock away WebSocket.close.
    monkeypatch.setattr(WebSocket, 'close', noop_close)

    # NOTE(vytas): Shorten the timeout so that we do not wait for 5 seconds.
    monkeypatch.setattr(
        testing.ASGIWebSocketSimulator, '_DEFAULT_WAIT_READY_TIMEOUT', 0.5
    )

    async with conductor as c:
        with pytest.raises(asyncio.TimeoutError):
            async with c.simulate_ws():
                pass


@pytest.mark.skipif(msgpack, reason='test requires msgpack lib to be missing')
def test_msgpack_missing():
    options = WebSocketOptions()
    handler = options.media_handlers[falcon.WebSocketPayloadType.BINARY]

    with pytest.raises(RuntimeError):
        handler.serialize({})

    with pytest.raises(RuntimeError):
        handler.deserialize(b'{}')


@pytest.mark.parametrize('reason', ['Client closing connection', '', None])
async def test_client_close_with_reason(reason, conductor):
    class Resource:
        def __init__(self):
            pass

        async def on_websocket(self, req, ws):
            await ws.accept()
            while True:
                try:
                    await ws.receive_data()

                except falcon.WebSocketDisconnected:
                    break

    resource = Resource()
    conductor.app.add_route('/', resource)

    async with conductor as c:
        async with c.simulate_ws('/', spec_version='2.3') as ws:
            await ws.close(4099, reason)

    assert ws.close_code == 4099
    if reason:
        assert ws.close_reason == reason
    else:
        assert ws.close_reason == ''


@pytest.mark.parametrize('reason', ['PEBCAK', 'wow such reason', '', None])
async def test_close_with_reason_no_cm(conductor, reason):
    class Resource:
        async def on_websocket(self, req, ws):
            await ws.accept()
            text = await ws.receive_text()
            await ws.send_text(text.upper())
            await ws.close(4001, reason)

    resource = Resource()
    conductor.app.add_route('/', resource)

    # NOTE(vytas): Here we don't use the async context manager pattern in order
    #   to collect coverage under CPython 3.11. It doesn't seem to help though.
    context = conductor.simulate_ws(spec_version='2.4')
    ws = context._ws

    await ws.wait_ready()
    await ws.send_text('Hello, World!')
    received = await ws.receive_text()
    assert received == 'HELLO, WORLD!'

    with pytest.raises(falcon.WebSocketDisconnected):
        await ws.receive_text()

    assert ws.close_reason == (reason or '')


@pytest.mark.parametrize('reason', ['PEBCAK', 'wow such reason', '', None])
@pytest.mark.parametrize('spec_version', ['2.2', '2.3', '2.4'])
async def test_close_with_reason(conductor, reason, spec_version):
    class Resource:
        async def on_websocket(self, req, ws):
            await ws.accept()
            await ws.close(3400, reason)

    resource = Resource()
    conductor.app.add_route('/', resource)

    async with conductor as c:
        async with c.simulate_ws('/', spec_version=spec_version) as ws:
            # Make sure the responder has a chance to reach the close() statement
            for _ in range(3):
                await asyncio.sleep(0)
            assert ws.closed
            assert ws.close_code == 3400

    assert ws.close_code == 3400
    if spec_version == '2.2':
        assert ws.close_reason == ''
    else:
        assert ws.close_reason == reason or 'Bad Request'


@pytest.mark.parametrize('no_default', [True, False])
@pytest.mark.parametrize(
    'code,expected',
    [
        (None, 'Normal Closure'),
        (1011, 'Internal Server Error'),
        (3405, 'Method Not Allowed'),
        (3701, ''),
        (3702, 'Emacs'),
        (4042, ''),
        (4099, 'wow such reason'),
    ],
)
async def test_reason_mapping(no_default, code, expected, conductor):
    class Resource:
        def __init__(self):
            pass

        async def on_websocket(self, req, ws):
            await ws.accept()
            await ws.close(code)

    resource = Resource()
    conductor.app.add_route('/', resource)
    if no_default:
        conductor.app.ws_options.default_close_reasons = {}
    else:
        # NOTE(vytas): Although it would be fun, we opt not to provide reasons
        #   for 7xx errors by default.
        conductor.app.ws_options.default_close_reasons[3702] = 'Emacs'
        conductor.app.ws_options.default_close_reasons[4099] = 'wow such reason'

    async with conductor as c:
        with pytest.raises(falcon.WebSocketDisconnected):
            async with c.simulate_ws('/', spec_version='2.10.3') as ws:
                await ws.receive_data()

    if code:
        assert ws.close_code == code
    else:
        assert ws.close_code == CloseCode.NORMAL

    if no_default:
        assert ws.close_reason == ''
    else:
        assert ws.close_reason == expected


@pytest.mark.parametrize('status', [200, 500, 422, 400])
@pytest.mark.parametrize('thing', [falcon.HTTPStatus, falcon.HTTPError])
@pytest.mark.parametrize('accept', [True, False])
async def test_ws_http_error_or_status_response(conductor, status, thing, accept):
    class Resource:
        async def on_websocket(self, req, ws):
            if accept:
                await ws.accept()
            raise thing(status)

    conductor.app.add_route('/', Resource())
    exp_code = 3000 + status

    async with conductor as c:
        if accept:
            async with c.simulate_ws() as ws:
                # Make sure the responder has a chance to reach the raise point
                for _ in range(3):
                    await asyncio.sleep(0)
                assert ws.closed
                assert ws.close_code == exp_code
        else:
            with pytest.raises(falcon.WebSocketDisconnected) as err:
                async with c.simulate_ws():
                    pass
            assert err.value.code == exp_code


@pytest.mark.parametrize('status', [200, 500, 422, 400])
@pytest.mark.parametrize(
    'thing',
    [
        falcon.HTTPStatus,
        falcon.HTTPError,
    ],
)
@pytest.mark.parametrize('place', ['request', 'resource'])
async def test_ws_http_error_or_status_middleware(conductor, status, thing, place):
    called = False

    class Resource:
        async def on_websocket(self, req, ws):
            nonlocal called
            called = True

    class Middleware:
        async def process_request_ws(self, req, ws):
            if place == 'request':
                raise thing(status)

        async def process_resource_ws(self, req, ws, res, params):
            if place == 'resource':
                raise thing(status)

    conductor.app.add_route('/', Resource())
    conductor.app.add_middleware(Middleware())
    exp_code = 3000 + status

    async with conductor as c:
        with pytest.raises(falcon.WebSocketDisconnected) as err:
            async with c.simulate_ws():
                pass
        assert err.value.code == exp_code
    assert not called


class FooBarError(Exception):
    pass


@pytest.mark.parametrize('status', [200, 500, 422, 400])
@pytest.mark.parametrize('thing', [falcon.HTTPStatus, falcon.HTTPError])
@pytest.mark.parametrize(
    'place', ['request', 'resource', 'ws_before_accept', 'ws_after_accept']
)
@pytest.mark.parametrize('handler_has_ws', [True, False])
async def test_ws_http_error_or_status_error_handler(
    conductor, status, thing, place, handler_has_ws
):
    class Resource:
        async def on_websocket(self, req, ws):
            if place == 'ws_before_accept':
                raise FooBarError
            await ws.accept()
            if place == 'ws_after_accept':
                raise FooBarError

    class Middleware:
        async def process_request_ws(self, req, ws):
            if place == 'request':
                raise FooBarError

        async def process_resource_ws(self, req, ws, res, params):
            if place == 'resource':
                raise FooBarError

    if handler_has_ws:

        async def handle_foobar(req, resp, ex, param, ws=None):
            raise thing(status)

    else:

        async def handle_foobar(req, resp, ex, param):  # type: ignore[misc]
            raise thing(status)

    conductor.app.add_route('/', Resource())
    conductor.app.add_middleware(Middleware())
    conductor.app.add_error_handler(FooBarError, handle_foobar)
    exp_code = 3000 + status

    async with conductor as c:
        if place == 'ws_after_accept':
            async with c.simulate_ws() as ws:
                # Make sure the responder has a chance to reach the raise point
                for _ in range(3):
                    await asyncio.sleep(0)
                assert ws.closed
                assert ws.close_code == exp_code
        else:
            with pytest.raises(falcon.WebSocketDisconnected) as err:
                async with c.simulate_ws():
                    pass
            assert err.value.code == exp_code


@pytest.mark.parametrize('max_receive_queue', [0, 1, 4, 7, 17])
async def test_max_receive_queue_sizes(conductor, max_receive_queue):
    class Chat:
        async def on_websocket(self, req, ws, user):
            await ws.accept()

            broadcast = [
                '[Guest123] ping',
                '[John F. Barbaz] Hi everyone.',
                f'Hello, {user}!',
            ]

            while True:
                await ws.send_text(broadcast.pop())

                msg = await ws.receive_text()
                if msg == '/quit':
                    await ws.send_text('Bye!')
                    await ws.close(reason='Received /quit')
                    break
                else:
                    await ws.send_text(f'[{user}] {msg}')

    conductor.app.ws_options.max_receive_queue = max_receive_queue
    conductor.app.add_route('/chat/{user}', Chat())

    received = []
    messages = ['/quit', 'I have to leave this test soon.', 'Hello!']

    async with conductor as c:
        async with c.simulate_ws('/chat/foobarer') as ws:
            while messages:
                received.append(await ws.receive_text())

                await ws.send_text(messages.pop())
                received.append(await ws.receive_text())

    assert len(received) == 6


@pytest.mark.parametrize('sink', ['no-ws-arg', 'with-ws-arg', 'unhashable-obj'])
async def test_sinks(conductor, sink):
    async def say_hello(req, resp_or_ws, name):
        greeting = f'Hello, {name}!'
        if req.is_websocket:
            await resp_or_ws.accept()
            await resp_or_ws.send_text(greeting)

            await resp_or_ws.receive_text()  # Receive /bye
        else:
            resp_or_ws.content_type = falcon.MEDIA_TEXT
            resp_or_ws.text = greeting

    async def sink_no_ws_arg(req, resp_or_ws, name):
        assert resp_or_ws is not None
        await say_hello(req, resp_or_ws, name)

    async def sink_with_ws_arg(req, resp, name, ws=None):
        if ws:
            assert resp is None
            await say_hello(req, ws, name)
        else:
            assert resp is not None
            await say_hello(req, resp, name)

    class Unhashable(list):
        async def __call__(self, req, resp, name, ws=None):
            await sink_with_ws_arg(req, resp, name, ws=ws)

    sinks = {
        'no-ws-arg': sink_no_ws_arg,
        'with-ws-arg': sink_with_ws_arg,
        'unhashable-obj': Unhashable(),
    }
    conductor.app.add_sink(sinks[sink], r'/ws/(?P<name>\w+)')

    async with conductor as c:
        resp = await c.simulate_get('/ws/World')
        assert resp.status_code == 200
        assert resp.text == 'Hello, World!'

        expectation = (
            pytest.warns(DeprecatedWarning) if sink == 'no-ws-arg' else nullcontext()
        )
        with expectation:
            async with c.simulate_ws('/ws/World') as ws:
                assert await ws.receive_text() == 'Hello, World!'
                await ws.send_text('/bye')


async def test_static_route(conductor, tmp_path):
    conductor.app.add_static_route('/static', tmp_path)

    async with conductor as c:
        with pytest.raises(falcon.WebSocketDisconnected) as exc_info:
            async with c.simulate_ws('/static/report.json'):
                pass

        assert exc_info.value.code == 3400



from __future__ import annotations

import asyncio
import logging
import typing
import uuid

from falcon.asgi import Request
from falcon.asgi import Response
from falcon.asgi import SSEvent
from falcon.asgi import WebSocket

logger = logging.getLogger(__name__)


class Emitter:
    POLL_TIMEOUT = 3.0

    def __init__(self) -> None:
        self._done: bool = False
        # NOTE(tang-vu): Bounded queue prevents unbounded memory growth when
        #   SSE consumers stop reading but remain connected. While unlikely in
        #   this short-lived E2E test, this guards against memory leaks in
        #   long-running scenarios, and demonstrates safe concurrency patterns.
        self._queue: asyncio.Queue[SSEvent] = asyncio.Queue(maxsize=32)

    async def events(self) -> typing.AsyncGenerator[SSEvent | None, None]:
        try:
            yield SSEvent(text='SSE CONNECTED')

            while True:
                try:
                    event = await asyncio.wait_for(
                        self._queue.get(), timeout=self.POLL_TIMEOUT
                    )
                    yield event
                except asyncio.TimeoutError:
                    # NOTE(vytas): Keep the connection alive.
                    yield None
        finally:
            # TODO(vytas): Is there a more elegant way to detect a disconnect?
            self._done = True

    async def enqueue(self, message: str) -> None:
        event = SSEvent(text=message, event_id=str(uuid.uuid4()))
        try:
            self._queue.put_nowait(event)
        except asyncio.QueueFull:
            logger.warning(
                'SSE emitter queue full (maxsize=%d), dropping message. '
                'This indicates a slow or stalled consumer.',
                self._queue.maxsize,
            )

    @property
    def done(self) -> bool:
        return self._done


class Hub:
    def __init__(self) -> None:
        self._emitters: set[Emitter] = set()
        self._users: dict[str, WebSocket] = {}

    def _update_emitters(self) -> set[Emitter]:
        done = {emitter for emitter in self._emitters if emitter.done}
        self._emitters.difference_update(done)
        return self._emitters.copy()

    def add_user(self, name: str, ws: WebSocket) -> None:
        self._users[name] = ws

    def remove_user(self, name: str) -> None:
        self._users.pop(name, None)

    async def broadcast(self, message: str) -> None:
        for emitter in self._update_emitters():
            await emitter.enqueue(message)

    async def message(self, name: str, text: str) -> None:
        ws = self._users.get(name)
        if ws:
            # TODO(vytas): What if this overlaps with another ongoing send?
            await ws.send_text(text)

    def events(self) -> typing.AsyncGenerator[SSEvent | None, None]:
        emitter = Emitter()
        self._update_emitters()
        self._emitters.add(emitter)
        return emitter.events()


class Events:
    def __init__(self, hub: Hub):
        self._hub = hub

    async def on_get(self, req: Request, resp: Response) -> None:
        resp.sse = self._hub.events()



# Copyright 2025 by Vytautas Liuolia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Information on stable Falcon releases.

This extension aggregates a table of maintained and EOL Falcon releases from
the provided path containing changelogs.
"""

import collections
import datetime
import enum
import itertools
import os
import pathlib
import re
import time

import sphinx.util.docutils

FALCON_DOCS_ROOT = pathlib.Path(__file__).resolve().parent.parent


class _FalconRelease:
    """A stable Falcon release."""

    _CHANGELOG_PATTERN = re.compile(r'^(\d+)\.(\d+).(\d+)\.rst$')
    _RELEASE_META_PATTERN = re.compile(r'\.\.\s+falcon-release\:\s+([\d-]+)')

    _PYPI_API_URL = 'https://pypi.org/pypi/falcon/{version}/json'

    @classmethod
    def from_changelog(cls, path):
        if not (mt := cls._CHANGELOG_PATTERN.match(path.name)):
            return None

        version = tuple(map(int, mt.groups()))
        if mt := cls._RELEASE_META_PATTERN.search(path.read_text()):
            date = datetime.date.fromisoformat(mt.group(1))
        else:
            date = None

        return cls(version, date)

    @classmethod
    def gather(cls, changelog_path):
        releases = [cls.from_changelog(path) for path in changelog_path.iterdir()]
        return sorted(
            release
            for release in releases
            if release is not None and not release.is_zerover
        )

    @property
    def is_zerover(self):
        return self.major == 0

    @property
    def major(self):
        return self.version[0]

    @property
    def version_str(self):
        return '.'.join(map(str, self.version))

    @property
    def version_major_minor(self):
        return '.'.join(map(str, self.version[:2]))

    def __init__(self, version, date=None):
        self.version = version
        self.date = date

    @property
    def doc_link(self):
        return f':doc:`{self.version_str} </changes/{self.version_str}>`'

    @property
    def doc_link_major_minor(self):
        return f':doc:`{self.version_major_minor} </changes/{self.version_str}>`'

    @property
    def rst_repr(self):
        return f'{self.doc_link} ({self.date})'

    def __eq__(self, other):
        return self.version == other.version

    def __lt__(self, other):
        return self.version < other.version

    def __repr__(self):
        return f'<falcon-{self.version_str} {self.date}>'

    def _load_date_from_pypi(self):
        # NOTE(vytas): This is not really used, but it is handy to verify the
        #   dates of historical releases, and other types of debugging.
        import requests

        resp = requests.get(self._PYPI_API_URL.format(version=self.version_str))
        if resp.status_code == 200:
            artifact, *_ = resp.json()['urls']
            upload_date, _, _ = artifact['upload_time'].partition('T')
            self.date = datetime.date.fromisoformat(upload_date)
        return self.date


_FalconMajorMinor = collections.namedtuple('_FalconMajorMinor', ('first', 'latest'))


class _ReleaseStatus(enum.Enum):
    current = 'current'
    security = 'security'
    eol = 'eol'

    @property
    def ref(self):
        return f':ref:`stable_release_{self.value}`'


class FalconReleasesDirective(sphinx.util.docutils.SphinxDirective):
    """Directive to tabulate the summary of stable Falcon releases."""

    _OLDSTABLE_MAINTENANCE_PERIOD = datetime.timedelta(days=365 * 2)

    required_arguments = 1
    has_content = True

    @staticmethod
    def get_build_date():
        # NOTE(vytas,bmwiedemann): Afford overriding the current date with
        #   SOURCE_DATE_EPOCH for reproducible documentation builds.
        #   (See also: https://reproducible-builds.org/docs/source-date-epoch/)
        build_date = datetime.datetime.fromtimestamp(
            int(os.environ.get('SOURCE_DATE_EPOCH', time.time())),
            tz=datetime.timezone.utc,
        )
        return build_date.date()

    def run(self):
        changelog_path = pathlib.Path(self.arguments[0])
        if not changelog_path.is_absolute():
            changelog_path = FALCON_DOCS_ROOT / changelog_path

        releases = _FalconRelease.gather(changelog_path)
        for release in releases[:-1]:
            assert release.date is not None, (
                f'{release} has no release date metadata (and is not latest)'
            )
        if releases[-1].date is None:
            releases.pop()

        stable_releases = []
        for _, group in itertools.groupby(
            reversed(releases), lambda rel: rel.version_major_minor
        ):
            series = list(group)
            stable_releases.append(
                _FalconMajorMinor(first=min(series), latest=max(series))
            )

        current = stable_releases[0]
        current_series = sorted(
            rel for rel in stable_releases if rel.first.major == current.first.major
        )
        oldstable = max(
            rel for rel in stable_releases if rel.first.major == current.first.major - 1
        )

        header = '.. list-table::\n    :header-rows: 1\n\n'
        rows = [
            (
                'Major/Minor Version',
                'First Release',
                'Latest Patch Release',
                'Release Status',
            )
        ]

        for first, latest in stable_releases:
            eol_label = ''
            if latest == current.latest:
                status = _ReleaseStatus.current
            elif latest == oldstable.latest:
                eol_date = (
                    current_series[0].first.date + self._OLDSTABLE_MAINTENANCE_PERIOD
                )
                if self.get_build_date() < eol_date:
                    status = _ReleaseStatus.security
                    # NOTE(vytas): Move to a new line because otherwise table
                    #   layout gets messed up in an unpleasant way.
                    eol_label = f'\n\n        (until {eol_date})'
                else:
                    status = _ReleaseStatus.eol
                    eol_label = f'(since {eol_date})'
            else:
                status = _ReleaseStatus.eol

            rows.append(
                (
                    first.doc_link_major_minor,
                    first.rst_repr,
                    latest.version_str if latest == first else latest.rst_repr,
                    f'{status.ref} {eol_label}',
                )
            )

        content = '\n\n'.join(
            '    * ' + '\n      '.join(f'- {cell}' for cell in row) for row in rows
        )

        return self.parse_text_to_nodes(header + content)


def setup(app):
    app.add_directive('falcon-releases', FalconReleasesDirective)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from urllib.parse import quote

import pytest

import falcon
from falcon import errors
from falcon import testing


class CaptureQueryStringMedia:
    """Resource that captures the deserialized query string media."""

    def on_get(self, req, resp):
        self.captured_media = req.get_query_string_as_media()

    def on_post(self, req, resp):
        self.captured_media = req.get_query_string_as_media('application/json')


@pytest.fixture()
def standard_client(asgi, util):
    """Fixture providing a test client with CaptureQueryStringMedia resource."""
    resource = CaptureQueryStringMedia()
    app = util.create_app(asgi)
    app.add_route('/test', resource)
    return testing.TestClient(app), resource


class TestQueryStringAsMedia:
    """Test query string deserialization as media."""

    def test_simple_json_query_string(self, standard_client):
        """Test deserializing a simple JSON query string."""
        client, resource = standard_client

        # Query string: {"key": "value"}
        json_data = '{"key": "value"}'
        query_string = quote(json_data, safe='')

        client.simulate_get('/test', query_string=query_string)

        assert resource.captured_media == {'key': 'value'}

    def test_complex_json_query_string(self, standard_client):
        """Test OpenAPI 3.2 example with complex JSON."""
        client, resource = standard_client

        # Query string: {"numbers":[1,2],"flag":null}
        json_data = '{"numbers":[1,2],"flag":null}'
        query_string = quote(json_data, safe='')

        client.simulate_get('/test', query_string=query_string)

        assert resource.captured_media == {'numbers': [1, 2], 'flag': None}

    def test_empty_query_string(self, asgi, util):
        """Test behavior with empty query string."""

        class ResourceWithDefault:
            def on_get(self, req, resp):
                self.captured_media = req.get_query_string_as_media(
                    default_when_empty={'default': 'value'}
                )

        resource = ResourceWithDefault()
        app = util.create_app(asgi)
        app.add_route('/test', resource)
        client = testing.TestClient(app)

        client.simulate_get('/test')

        # Should return the default value when empty and handler raises error
        assert resource.captured_media == {'default': 'value'}

    def test_invalid_json_query_string(self, asgi, util):
        """Test error handling with invalid JSON."""

        class ResourceInvalidJSON:
            def on_get(self, req, resp):
                try:
                    req.get_query_string_as_media()
                except errors.HTTPBadRequest:
                    self.error_caught = True
                else:
                    self.error_caught = False

        resource = ResourceInvalidJSON()
        app = util.create_app(asgi)
        app.add_route('/test', resource)
        client = testing.TestClient(app)

        # Invalid JSON
        invalid_json = '{"incomplete"'
        query_string = quote(invalid_json, safe='')

        client.simulate_get('/test', query_string=query_string)

        assert resource.error_caught

    def test_explicit_media_type(self, standard_client):
        """Test specifying an explicit media type."""
        client, resource = standard_client

        json_data = '{"explicit": "type"}'
        query_string = quote(json_data, safe='')

        client.simulate_post('/test', query_string=query_string)

        assert resource.captured_media == {'explicit': 'type'}

    def test_special_characters_in_json(self, standard_client):
        """Test JSON with special characters."""
        client, resource = standard_client

        # JSON with special characters
        json_data = '{"name": "Test & Demo", "value": "100%"}'
        query_string = quote(json_data, safe='')

        client.simulate_get('/test', query_string=query_string)

        assert resource.captured_media == {'name': 'Test & Demo', 'value': '100%'}

    def test_nested_json_structures(self, standard_client):
        """Test deeply nested JSON structures."""
        client, resource = standard_client

        json_data = '{"level1": {"level2": {"level3": ["a", "b", "c"]}}}'
        query_string = quote(json_data, safe='')

        client.simulate_get('/test', query_string=query_string)

        expected = {'level1': {'level2': {'level3': ['a', 'b', 'c']}}}
        assert resource.captured_media == expected

    def test_different_media_types(self, asgi, util):
        """Test with different media type handlers."""

        class ResourceCustomMediaType:
            def on_get(self, req, resp):
                # Plain text - should just return the decoded string
                self.captured_media = req.get_query_string_as_media(
                    'application/x-www-form-urlencoded'
                )

        resource = ResourceCustomMediaType()
        app = util.create_app(asgi)
        app.add_route('/test', resource)
        client = testing.TestClient(app)

        # URL-encoded form data
        query_string = quote('key1=value1&key2=value2', safe='')

        client.simulate_get('/test', query_string=query_string)

        # URLEncodedFormHandler should parse this
        assert resource.captured_media == {'key1': 'value1', 'key2': 'value2'}

    def test_unicode_in_query_string(self, standard_client):
        """Test JSON with Unicode characters."""
        client, resource = standard_client

        # JSON with Unicode
        json_data = '{"emoji": "🚀", "chinese": "你好"}'
        query_string = quote(json_data, safe='')

        client.simulate_get('/test', query_string=query_string)

        assert resource.captured_media == {'emoji': '🚀', 'chinese': '你好'}

    def test_error_caching(self, asgi, util):
        """Test error behavior on repeated calls."""

        class ResourceErrorCaching:
            def __init__(self):
                self.call_count = 0

            def on_get(self, req, resp):
                self.call_count += 1
                try:
                    req.get_query_string_as_media()
                except errors.HTTPBadRequest:
                    pass

        resource = ResourceErrorCaching()
        app = util.create_app(asgi)
        app.add_route('/test', resource)
        client = testing.TestClient(app)

        invalid_json = '{"bad"'
        query_string = quote(invalid_json, safe='')

        # First call
        client.simulate_get('/test', query_string=query_string)
        # Second call
        client.simulate_get('/test', query_string=query_string)

        # Both calls should have been made
        assert resource.call_count == 2

    def test_default_when_empty_not_used_for_valid_data(self, asgi, util):
        """Test that default_when_empty is not used when data is valid."""

        class ResourceWithDefault:
            def on_get(self, req, resp):
                self.captured_media = req.get_query_string_as_media(
                    default_when_empty={'should': 'not see this'}
                )

        resource = ResourceWithDefault()
        app = util.create_app(asgi)
        app.add_route('/test', resource)
        client = testing.TestClient(app)

        json_data = '{"actual": "data"}'
        query_string = quote(json_data, safe='')

        client.simulate_get('/test', query_string=query_string)

        # Should get the actual data, not the default
        assert resource.captured_media == {'actual': 'data'}

    def test_error_propagation(self, asgi, util):
        """Test that non-MediaNotFoundError exceptions propagate correctly."""

        class FailingHandler:
            exhaust_stream = False

            def deserialize(self, stream, content_type, content_length):
                raise ValueError('Custom error')

        class ResourceErrorCheck:
            def on_get(self, req, resp):
                try:
                    req.get_query_string_as_media('application/custom')
                except ValueError as e:
                    self.error_message = str(e)

        resource = ResourceErrorCheck()
        handler = FailingHandler()

        app = util.create_app(asgi)
        app.req_options.media_handlers['application/custom'] = handler
        app.add_route('/test', resource)
        client = testing.TestClient(app)

        client.simulate_get('/test', query_string='data')

        assert resource.error_message == 'Custom error'

    def test_cached_error_with_default_when_empty(self, asgi, util):
        """Test that an error followed by default_when_empty returns default."""

        class ResourceCachedError:
            def on_get(self, req, resp):
                # First call - will fail
                try:
                    req.get_query_string_as_media()
                except Exception:
                    pass
                # Second call - should return default
                self.result = req.get_query_string_as_media(default_when_empty={})

        resource = ResourceCachedError()
        app = util.create_app(asgi)
        app.add_route('/test', resource)
        client = testing.TestClient(app)

        # Empty query string will cause MediaNotFoundError
        client.simulate_get('/test', query_string='')

        assert resource.result == {}

    def test_cached_error_reraises_without_default(self, asgi, util):
        """Test that error is re-raised on subsequent calls."""

        class ResourceCachedErrorReraise:
            def on_get(self, req, resp):
                # First call - will fail
                try:
                    req.get_query_string_as_media()
                except falcon.MediaNotFoundError:
                    pass
                # Second call - should re-raise error
                try:
                    req.get_query_string_as_media()
                except falcon.MediaNotFoundError as err:
                    self.error_caught = str(err)

        resource = ResourceCachedErrorReraise()
        app = util.create_app(asgi)
        app.add_route('/test', resource)
        client = testing.TestClient(app)

        client.simulate_get('/test', query_string='')

        assert 'MediaNotFoundError' in resource.error_caught

    def test_unsupported_media_type_raises_value_error(self, asgi, util):
        """Test unsupported media type raises ValueError."""

        class ResourceUnsupportedMedia:
            def on_get(self, req, resp):
                try:
                    # Request a media type that doesn't exist
                    req.get_query_string_as_media('application/x-nonexistent')
                except ValueError as e:
                    self.error_caught = True
                    self.error_message = str(e)
                except Exception:
                    self.error_caught = False

        resource = ResourceUnsupportedMedia()
        app = util.create_app(asgi)
        app.add_route('/test', resource)
        client = testing.TestClient(app)

        json_data = '{"test": "data"}'
        query_string = quote(json_data, safe='')

        client.simulate_get('/test', query_string=query_string)

        assert resource.error_caught
        assert 'No media handler is configured' in resource.error_message
        assert 'application/x-nonexistent' in resource.error_message
